{******************************************************************************}
{                     -                         }
{                                                                              }
{                                                                 }
{                    Copyright (c) 2001 BaseGroup Labs                         }
{******************************************************************************}

unit WavUtils;

{$DEFINE UseExceptions} //          

interface

type
  TVector = array of Double; //     

  TDecomposition = record      //   
    WaveletOrder: Integer;     //  
    SignalLength: Integer;     //   
    Approx: TVector;           //      
    Details: array of TVector; //      
                               //     
  end;

{$IFNDEF UseExceptions}
//  ,   
type
  TWaveletResult = (
    wavSuccess,             //   
    wavErrorWrongOrder,     //    
    wavErrorSignalTooShort, //     
    wavErrorWrongLength     //       
  );
{$ENDIF}

//      
resourcestring
  SWrongOrder = '  ';
  SSignalTooShort = '       ';
  SWrongLength = '     ';

//      .
//     Signal  
//    Order   Level.
//   - ,  
//      ,
//      ,
//      
// (   ).
function Decompose(
  const Signal: TVector;      //  
  const Order: Integer;       //  
  const Level: Integer        //  
  {$IFNDEF UseExceptions};
  var Decomp : TDecomposition //  
  ): TWaveletResult;          //  
  {$ELSE}
  ): TDecomposition;          //  
  {$ENDIF}

//  .
//      Decomp.
//   -   .
//       
//  .
function Reconstruct(
  const Decomp: TDecomposition //  
  {$IFNDEF UseExceptions};
  var SynthSig: TVector        //  
  ): TWaveletResult;           //  
  {$ELSE}
  ): TVector;                  //  
  {$ENDIF}

//       -.
//      Signal  
//    Order.   - ,
//       Level.
function Denoise(
  const Signal: TVector; //  
  const Order: Integer;  //  
  const Level: Integer   //  
  {$IFNDEF UseExceptions};
  var SynthSig: TVector  //  
  ): TWaveletResult;     //  
  {$ELSE}
  ): TVector;            //  
  {$ENDIF}
  overload;

//       -.
//     
//    .
function Denoise(
  const Decomp: TDecomposition //   
  {$IFNDEF UseExceptions};
  var SynthSig: TVector        //  
  ): TWaveletResult;           //  
  {$ELSE}
  ): TVector;                  //  
  {$ENDIF}
  overload;

implementation

uses SysUtils;

type
  PDouble = ^Double;

const
  NWavelets = 10;

  Daubechies1: array [0..1] of Double =
  (  0.50000000000000,  0.50000000000000 );

  Daubechies2: array [0..3] of Double =
  (  0.34150635094622,  0.59150635094587,  0.15849364905378, -0.09150635094587 );

  Daubechies3: array [0..5] of Double =
  (  0.23523360389270,  0.57055845791731,  0.32518250026371, -0.09546720778426,
    -0.06041610415535,  0.02490874986589 );

  Daubechies4: array [0..7] of Double =
  (  0.16290171402562,  0.50547285754565,  0.44610006912319, -0.01978751311791,
    -0.13225358368437,  0.02180815023739,  0.02325180053556, -0.00749349466513 );

  Daubechies5: array [0..9] of Double =
  (  0.11320949129173,  0.42697177135271,  0.51216347213016,  0.09788348067375,
    -0.17132835769133, -0.02280056594205,  0.05485132932108, -0.00441340005433,
    -0.00889593505093,  0.00235871396920 );

  Daubechies6: array [0..11] of Double =
  (  0.07887121600143,  0.34975190703757,  0.53113187994121,  0.22291566146505,
    -0.15999329944587, -0.09175903203003,  0.06894404648720,  0.01946160485396,
    -0.02233187416548,  0.00039162557603,  0.00337803118151, -0.00076176690258 );

  Daubechies7: array [0..13] of Double =
  (  0.05504971537285,  0.28039564181304,  0.51557424581833,  0.33218624110566,
    -0.10175691123173, -0.15841750564054,  0.05042323250485,  0.05700172257986,
    -0.02689122629486, -0.01171997078235,  0.00887489618962,  0.00030375749776,
    -0.00127395235906,  0.00025011342658 );

  Daubechies8: array [0..15] of Double =
  (  0.03847781105406,  0.22123362357624,  0.47774307521438,  0.41390826621166,
    -0.01119286766665, -0.20082931639111,  0.00033409704628,  0.09103817842345,
    -0.01228195052300, -0.03117510332533,  0.00988607964808,  0.00618442240954,
    -0.00344385962813, -0.00027700227421,  0.00047761485533, -0.00008306863060 );

  Daubechies9: array [0..17] of Double =
  (  0.02692517479416,  0.17241715192471,  0.42767453217028,  0.46477285717278,
     0.09418477475112, -0.20737588089628, -0.06847677451090,  0.10503417113714,
     0.02172633772990, -0.04782363205882,  0.00017744640673,  0.01581208292614,
    -0.00333981011324, -0.00302748028715,  0.00130648364018,  0.00016290733601,
    -0.00017816487955,  0.00002782275679 );

  Daubechies10: array [0..19] of Double =
  (  0.01885857879640,  0.13306109139687,  0.37278753574266,  0.48681405536610,
     0.19881887088440, -0.17666810089647, -0.13855493935993,  0.09006372426666,
     0.06580149355070, -0.05048328559801, -0.02082962404385,  0.02348490704841,
     0.00255021848393, -0.00758950116768,  0.00098666268244,  0.00140884329496,
    -0.00048497391996, -0.00008235450295,  0.00006617718320, -0.00000937920789 );

  Daubechies: array [1..NWavelets] of PDouble =
  ( @Daubechies1[0], @Daubechies2[0], @Daubechies3[0], @Daubechies4[0],
    @Daubechies5[0], @Daubechies6[0], @Daubechies7[0], @Daubechies8[0],
    @Daubechies9[0], @Daubechies10[0] );

var
  WaveletsInfo: array [1..NWavelets] of TVector;

function Max(A,B: Integer): Integer;
begin
  if A > B then Result := A else Result := B;
end;

//   -.
//   -     ,   - 
//  (  ).   
//       (  ).
//    ,    .
{$IFNDEF UseExceptions}
function DWT(const Signal, Filter: TVector; var Approx, Details: TVector): TWaveletResult;
{$ELSE}
procedure DWT(const Signal, Filter: TVector; var Approx, Details: TVector);
{$ENDIF}
var
  Buffer: TVector;
  I, I0, J, J0, J1, J2, K, L, N: Integer;
begin
  //     
  K := Length(Signal); L := Length(Filter);
  //   
  N := K; if Odd(N) then Inc(N);
  //      
  if L > N then
  {$IFDEF UseExceptions}
    raise Exception.Create(SSignalTooShort);
  {$ELSE}
  begin
    Result := wavErrorSignalTooShort;
    Exit;
  end;
  Result := wavSuccess;
  {$ENDIF}
  //    
  SetLength(Buffer,N+L-2);
  //    
  Move(Signal[0],Buffer[0],K*SizeOf(Signal[0]));
  //    
  if Odd(K) then Buffer[K] := (Signal[K-1]+Signal[0])/2;
  //   
  if L > 2 then Move(Signal[0],Buffer[N],(L-2)*SizeOf(Signal[0]));
  //      ,    
  //   .     ,
  //         .
  Approx := nil; Details := nil;
  //     
  N := N div 2; K := L div 2;
  SetLength(Approx,N);
  SetLength(Details,N);
  //     
  for I := 0 to Pred(N) do //    
  begin
    I0 := I * 2;
    Approx[I] := 0; Details[I] := 0;
    for J := 0 to Pred(K) do //    
    begin
      J0 := J * 2; J1 := J0 + 1; J2 := J0 + 2;
      Approx[I] := Approx[I] + Buffer[I0+J0]*Filter[J0] + Buffer[I0+J1]*Filter[J1];
      Details[I] := Details[I] + Buffer[I0+J0]*Filter[L-J1] - Buffer[I0+J1]*Filter[L-J2];
    end;
  end;
end;

//   -.
//   -      ( 
// ),   -  .  
//       
// (  ).      ,
//      OddLength  True.
//    ,    .
function IDWT(const Filter, Approx, Details: TVector; OddLength: Boolean = False): TVector;
  function FillBuf(const Source: TVector; N, Nb, L: Integer): TVector;
  begin
    SetLength(Result,L+N);
    if Nb > 0 then Move(Source[0],Result[L],Nb*SizeOf(Source[0]));
    if Nb < N then FillChar(Result[L+Nb],(N-Nb)*SizeOf(Source[0]),0);
    if L > 0 then Move(Result[N],Result[0],L*SizeOf(Source[0]));
  end;
var
  AppBuf, DetBuf: TVector;
  I, I0, I1, J, J0, J1, J2, K, L, N, Na, Nd: Integer;
begin
  //  
  AppBuf := nil; DetBuf := nil;
  //      .
  //     .
  Na := Length(Approx); Nd := Length(Details);
  //       .
  L := Length(Filter); K := Pred(L div 2); N := Max(Na,Nd);
  //    .
  AppBuf := FillBuf(Approx,N,Na,K);
  //    .
  DetBuf := FillBuf(Details,N,Nd,K);
  //     ( ).
  //      ,    !
  SetLength(Result,N*2);
  //      
  for I := 0 to Pred(N) do //    
  begin
    I0 := I * 2; I1 := I0 + 1;
    Result[I0] := 0; Result[I1] := 0;
    for J := 0 to K do //    
    begin
      J0 := J * 2; J1 := J0 + 1; J2 := J0 + 2;
      Result[I0] := Result[I0] + AppBuf[I+J]*Filter[L-J2] + DetBuf[I+J]*Filter[J1];
      Result[I1] := Result[I1] + AppBuf[I+J]*Filter[L-J1] - DetBuf[I+J]*Filter[J0];
    end;
  end;
  if OddLength then SetLength(Result,Pred(Length(Result)));
end;

//      .
function Decompose(
  const Signal: TVector;      //  
  const Order: Integer;       //  
  const Level: Integer        //  
{$IFNDEF UseExceptions};
  var Decomp : TDecomposition //  
  ): TWaveletResult;          //  
{$ELSE}
  ): TDecomposition;          //  
{$ENDIF}
var
  Filter: TVector;
  I: Integer;
begin
  Filter := nil;
  if (Order < Low(WaveletsInfo)) or (Order > High(WaveletsInfo)) then
  {$IFDEF UseExceptions}
    raise Exception.Create(SWrongOrder);
  {$ELSE}
  begin
    Result := wavErrorWrongOrder;
    Exit;
  end;
  Result := wavSuccess;
  {$ENDIF}
  Filter := WaveletsInfo[Order];
  with {$IFDEF UseExceptions}Result{$ELSE}Decomp{$ENDIF} do
  begin
    WaveletOrder := Length(Filter) div 2;
    SignalLength := Length(Signal);
    Approx := Signal;
    SetLength(Details,Level);
    FillChar(Details[0],Length(Details)*SizeOf(Details[0]),0);
    for I := 0 to Pred(Level) do
    {$IFDEF UseExceptions}
      DWT(Approx,Filter,Approx,Details[I]);
    {$ELSE}
    begin
      Result := DWT(Approx,Filter,Approx,Details[I]);
      if Result <> wavSuccess then Exit;
    end;
    {$ENDIF}
  end;
end;

//  .
function Reconstruct(
  const Decomp: TDecomposition //  
{$IFNDEF UseExceptions};
  var SynthSig: TVector        //  
  ): TWaveletResult;           //  
{$ELSE}
  ): TVector;                  //  
{$ENDIF}
var
{$IFDEF UseExceptions}
  SynthSig: TVector absolute Result;
{$ENDIF}
  Filter: TVector;
  I: Integer;
  IsOdd: Boolean;
begin
  Filter := nil;
  with Decomp do
  begin
    if (WaveletOrder < Low(WaveletsInfo)) or (WaveletOrder > High(WaveletsInfo)) then
    {$IFDEF UseExceptions}
      raise Exception.Create(SWrongOrder);
    {$ELSE}
    begin
      Result := wavErrorWrongOrder;
      Exit;
    end;
    {$ENDIF}
    Filter := WaveletsInfo[WaveletOrder];
    SynthSig := Approx;
    for I := High(Details) downto 0 do
    begin
      IsOdd := ((I > 0) and Odd(Length(Details[Pred(I)]))) or
               ((I = 0) and Odd(SignalLength));
      SynthSig := IDWT(Filter,SynthSig,Details[I],IsOdd);
    end;
    if SignalLength <> Length(SynthSig) then
    {$IFDEF UseExceptions}
      raise Exception.Create(SWrongLength);
    {$ELSE}
    begin
      Result := wavErrorWrongLength;
      Exit;
    end;
    Result := wavSuccess;
    {$ENDIF}
  end;
end;

//       -.
//       .
function Denoise(
  const Decomp: TDecomposition //   
  {$IFNDEF UseExceptions};
  var SynthSig: TVector        //  
  ): TWaveletResult;           //  
  {$ELSE}
  ): TVector;                  //  
  {$ENDIF}
  overload;
var
  I: Integer;
begin
  with Decomp do
  begin
    for I := 0 to High(Details) do
    begin
      FillChar(Details[I,0],Length(Details[I])*SizeOf(Details[I,0]),0);
    end;
  end;
  {$IFDEF UseExceptions}
  Result := Reconstruct(Decomp);
  {$ELSE}
  Result := Reconstruct(Decomp,SynthSig);
  {$ENDIF}
end;

//       -.
//      Signal  
//    Order.   - ,
//       Level.
function Denoise(
  const Signal: TVector; //  
  const Order: Integer;  //  
  const Level: Integer   //  
  {$IFNDEF UseExceptions};
  var SynthSig: TVector  //  
  ): TWaveletResult;     //  
  {$ELSE}
  ): TVector;            //  
  {$ENDIF}
  overload;
{$IFNDEF UseExceptions}
var
  Decomp: TDecomposition;
{$ENDIF}
begin
  {$IFDEF UseExceptions}
  Result := Denoise(Decompose(Signal,Order,Level));
  {$ELSE}
  Result := Decompose(Signal,Order,Level,Decomp);
  if Result <> wavSuccess then Exit;
  Result := Denoise(Decomp,SynthSig);
  {$ENDIF}
end;

var
  I, J: Integer;
  P: PDouble;

initialization
  for I := 1 to NWavelets do
  begin
    SetLength(WaveletsInfo[I],2*I);
    P := Daubechies[I];
    for J := 0 to Pred(2*I) do
    begin
      WaveletsInfo[I][J] := P^ * Sqrt(2);
      Inc(P);
    end;
  end;
end.
